import com.cattura.agents.configuration_module.model.interfaces.Plugin;
import com.cattura.agents.configuration_module.model.interfaces.Plugin.Inject;
import com.cattura.agents.configuration_module.model.interfaces.Plugin.On;
import com.cattura.agents.configuration_module.model.vos.plugin.PluginInfo;
import com.cattura.agents.configuration_module.model.interfaces.Observer.Observes;
import java.io.File;
import org.puremvc.java.multicore.patterns.facade.Facade
import com.cattura.agents.configuration_module.model.proxies.EnvironmentProxy
import com.cattura.agents.configuration_module.model.proxies.SettingsProxy
import com.cattura.agents.configuration_module.model.proxies.capture.MediaPackageProxy
import com.cattura.agents.configuration_module.model.vos.PublishRequest
import com.cattura.agents.configuration_module.model.vos.events.publish.PublishCompleteEvent
import com.cattura.share.processor.data_provider.elements.Metadata
import com.cattura.share.utils.data_structures.Either
import com.cattura.agents.bridge.model.GraphBuilder.Graph
import com.cattura.agents.bridge.model.GraphBuilder.Graph.Node
import com.cattura.agents.configuration_module.model.properties.PluginNames
import com.google.common.base.Optional
import com.google.common.base.Predicate
import com.cattura.agents.configuration_module.model.properties.SettingNames
import com.cattura.share.utils.settings.Settings
import com.cattura.agents.configuration_module.model.vos.capture.MediaPackage;
import com.cattura.agents.configuration_module.model.vos.events.Event
import com.cattura.agents.configuration_module.model.vos.events.MediaPackageEvent
import com.cattura.agents.configuration_module.model.vos.events.annotations.Internal
import com.cattura.cattura_commons.utilities.io.HTTP
import com.cattura.share.utils.json.parsing.JSON
import java.util.logging.Level
import com.google.common.collect.Maps
import com.cattura.share.processor.Processor.Factory.Service
import com.cattura.agents.configuration_module.controller.facade.ApplicationFacade
import com.cattura.agents.configuration_module.model.vos.events.EventType
import com.cattura.agents.configuration_module.model.vos.events.internal.TemplateConstructionEvent
import com.google.common.cache.CacheBuilder
import com.google.common.cache.Cache
import com.cattura.integration.learning_engine.LearningEngineClient
import com.cattura.integration.learning_engine.vos.Request
import com.cattura.agents.configuration_module.model.proxies.integration.OAuthProxy
import com.cattura.agents.configuration_module.model.vos.events.internal.PublishRequestConstructionEvent

class LearningEnginePlugin implements Plugin
{
    public static @Inject.Info PluginInfo _info;
    public static @Inject.Facade Facade _facade;
    private final HTTP _http = new HTTP.Default()
    private final JSON _json = new JSON.Default()
    
    private final Cache<String, LearningEngineClient> _clients = CacheBuilder.newBuilder().build();
    
    private static class Options
    {
        public static final String LEARNING_ENGINE = "learning-engine"
        public static final String COURSE_ID = "courseID"
        public static final String LE_ENTRY_ID = "entryID"
        public static final String LE_ENTRY_URL = "entryURL"
        public static final String LE_FAILURE = "failure";
    }
    
    private static class LESettingNames
    {
        public static final String LEARNING_ENGINE_URL = "learning-engine-url"
    }
    
    private static class Identifiers
    {
        public static final String LEARNING_ENGINE_CONFIG = "learning-engine-config"
    }
    
    private static class Defaults
    {
        public static final String LEARNING_ENGINE_URL = "https://le.catturavideo.com/"
    }    
    
    @On.Load
    public void onLoad ()
    {
        _info.getLogger().info("Learning Engine Plugin started")
    }
    
    /*
     * Called when creating a template. Checks for the LE options and 
     * attaches LE node if it exists
     */    
    @Observes(event=TemplateConstructionEvent.class)
    public void onLectureTemplateEvent (def $event)
    {
        def options = $event.getOptions()
        final Graph graph = $event.getGraph()

        if (options.containsKey(Options.LEARNING_ENGINE))
        {
            final Metadata leMetadata = new Metadata()
            leMetadata.setID(Identifiers.LEARNING_ENGINE_CONFIG)
            leMetadata.setName("Learning Engine metadata")
            
            options.each({ key, value ->
                if (value && key.startsWith("${Options.LEARNING_ENGINE}-"))
                {
                    leMetadata.put(key.substring(Options.LEARNING_ENGINE.length() + 1), value)
                }
            });
        
            graph.add(PluginNames.METADATA)
                 .linkParents(0)
                 .configure([
                    metadata: leMetadata
                 ])
        }
    }
    
    /*
     * Called whenever a PublishRequest is constructed from a CaptureRequest.
     * Will pass Learning Engine metadata from the graph/template into the PublishRequest.
     * Is not called when a PublishRequest is created by the user (e.g. for republishing).
     */
    @Observes(event=PublishRequestConstructionEvent.class)
    public void onPublishRequestConstructionEvent (final PublishRequestConstructionEvent $event)
    {
        final Graph graph = $event.graph
        final Node node = findLearningEngineNode(graph)
        if (!node)
        {
            return
        }
        
        _info.getLogger().info("Preparing Learning Engine publish for ${$event.captureRequest.title} to Learning Engine")        
        final PublishRequest request = $event.publishRequest
        final Map options = Maps.newLinkedHashMap(request.options)
        options[Options.LEARNING_ENGINE] = node.config().metadata.getMap()
        request.setOptions(options)
    }
    
    /*
     * Handles onPublishComplete event
     */
    @Observes(event=PublishCompleteEvent.class)
    public void onPublishCompleteEvent (final PublishCompleteEvent $event)
    {
        final MediaPackage mediaPackage = $event.getMediaPackage()
        if (!mediaPackage)
        {
            _info.getLogger().warning("MediaPackage was null! Cannot publish it to Learning Engine.")
            return
        }

        try
        {
            final PublishRequest request = $event.getPublishRequest()
            if (request)
            {
                final def leOptions = request.getOptions()[Options.LEARNING_ENGINE]
                if (leOptions == null)
                {
                    _info.getLogger().info("Not publishing ${mediaPackage.title} to Learning Engine because it is not configured for Learning Engine")
                    return
                }
                
                publishMediaPackage(mediaPackage, leOptions)
            }
            else
            {
                final Graph graph = mediaPackage.getGraph().get()
                final Node node = findLearningEngineNode(graph)
                if (!node)
                {
                    _info.getLogger().info("Not publishing ${mediaPackage.title} to Learning Engine because it is not configured for Learning Engine")
                    return
                }

                publishMediaPackage(mediaPackage, node.config().metadata.getMap())
            }
        }
        catch (e)
        {
            _info.getLogger().log(Level.INFO, "Error publishing ${mediaPackage.title} to Learning Engine: ${e.message}", e)
        }
    }
    
    public def publishMediaPackage (final MediaPackage $mediaPackage, final Map $properties = [:])
    {
        try
        {
            final String token = getLearningEngineToken()
            final String url = getLearningEngineURL()
            final Graph graph = $mediaPackage.getGraph().get()
            
            if (!token)
            {
                _info.getLogger().info("Not publishing ${$mediaPackage.title} to Learning Engine because no token is specified")
                return
            }
            else if(!url)
            {
                _info.getLogger().info("Not publishing ${$mediaPackage.title} to Learning Engine because the URL is not valid.")
                return
            }

            def client = getClient()
            def request = new Request($mediaPackage)

            def courseIds = []
            if ($properties.containsKey(Options.COURSE_ID))
            {
                courseIds = $properties[Options.COURSE_ID] instanceof Collection ? $properties[Options.COURSE_ID].collect({
                    try
                    {
                        return it instanceof String ? Integer.parseInt(it) : it                    
                    }
                    catch (final NumberFormatException $nfe)
                    {
                        return -1
                    }
                }) : [$properties[Options.COURSE_ID]];

                request = request.withCourses(getCourses().findAll({
                    return courseIds.contains(it.id)
                }))
            }

            def result = client.submit(request).get()
            saveLEMetadata($mediaPackage, [
                (Options.LE_ENTRY_ID): result.id,
                (Options.LE_ENTRY_URL): result.viewURL.get(),
                (Options.COURSE_ID): courseIds
            ])        

            _info.getLogger().info("Media package ${$mediaPackage.title} was published to Learning Engine (with courses ${request.courses}) as id ${result.id} with thumbnail ${result.thumbnail.or('<none>')}")
            _facade.sendNotification(ApplicationFacade.EMIT_EVENT, new LearningEnginePublishEvent($mediaPackage))
            return result            
        }
        catch (final Exception $exception)
        {
            saveLEMetadata($mediaPackage, [
                (Options.LE_FAILURE): $exception.message
            ])
            
            _facade.sendNotification(ApplicationFacade.EMIT_EVENT, new LearningEnginePublishFailureEvent($mediaPackage, $exception.message))
            throw $exception
        }
    }
    
    @Service
    public def publishMediaPackage (@Service.Parameter(name="mediaPackageId") String $mediaPackageId)
    {
        try
        {
            final MediaPackageProxy mediaPackageProxy = _facade.retrieveProxy(MediaPackageProxy.NAME)
            final MediaPackage mediaPackage = mediaPackageProxy.getMediaPackage($mediaPackageId)
            return publishMediaPackage(mediaPackage, [:])
        }
        catch (final Exception $e)
        {
            $e.printStackTrace()
            _info.getLogger().info("Could not publish ${$mediaPackageId} to Learning Engine: ${$e}")
            return false
        }
    }
    
    @Service
    public def publishMediaPackage (@Service.Parameter(name="mediaPackageId") String $mediaPackageId, @Service.Parameter(name="properties") Map $properties)
    {
        try
        {
            final MediaPackageProxy mediaPackageProxy = _facade.retrieveProxy(MediaPackageProxy.NAME)
            final MediaPackage mediaPackage = mediaPackageProxy.getMediaPackage($mediaPackageId)
            return publishMediaPackage(mediaPackage, $properties)
        }
        catch (final Exception $e)
        {
            $e.printStackTrace()
            _info.getLogger().info("Could not publish ${$mediaPackageId} to Learning Engine: ${$e}")
            return false
        }
    }
    
    @Service
    public def publishMediaPackage (@Service.Parameter(name="mediaPackageId") String $mediaPackageId, @Service.Parameter(name="courseId") int $courseId)
    {
        try
        {
            final MediaPackageProxy mediaPackageProxy = _facade.retrieveProxy(MediaPackageProxy.NAME)
            final MediaPackage mediaPackage = mediaPackageProxy.getMediaPackage($mediaPackageId)
            final Map properties = [:]
            properties[Options.COURSE_ID] = $courseId
            return publishMediaPackage(mediaPackage, properties)
        }
        catch (final Exception $e)
        {
            $e.printStackTrace()
            _info.getLogger().info("Could not publish ${$mediaPackageId} to Learning Engine: ${$e}")
            return false
        }
    }    
    
    @Service
    public def getCourses ()
    {
        try
        {
            return getClient().getCourses().get()
        }
        catch (final Exception $e)
        {
            _info.getLogger().info("Could not retrieve courses for Learning Engine: ${$e}")
            return []
        }
    }
    
    @Service
    public def getCourses (@Service.Parameter(name="memberId") int $memberId)
    {
        try
        {
            return getClient().getCourses(getMembers().find({
                return it.id == $memberId
            })).get()
        }
        catch (final Exception $e)
        {
            _info.getLogger().info("Could not retrieve courses for Learning Engine: ${$e}")
            return []
        }
    }
    
    @Service
    public def getMembers ()
    {
        try
        {        
            return getClient().getMembers().get()
        }
        catch (final Exception $e)
        {
            _info.getLogger().info("Could not retrieve members for Learning Engine: ${$e}")
            return []
        }
    }
    
    @Service
    public def publishMediaPackage (@Service.Parameter(name="mediaPackageId") String $mediaPackageId, @Service.Parameter(name="courseIds") List $courseIds)
    {
        try
        {
            final MediaPackageProxy mediaPackageProxy = _facade.retrieveProxy(MediaPackageProxy.NAME)
            final MediaPackage mediaPackage = mediaPackageProxy.getMediaPackage($mediaPackageId)
            final Map properties = [:]
            properties[Options.COURSE_ID] = $courseIds
            return publishMediaPackage(mediaPackage, properties)
        }
        catch (final Exception $e)
        {
            $e.printStackTrace()
            _info.getLogger().info("Could not publish ${$mediaPackageId} to Learning Engine: ${$e}")
            return false
        }
    }
    
    @Service
    public boolean isLearningEngineEnabled ()
    {
        /* 14 October 2016: The old implementation used the portal
         *     to check if the API token has access to the Learning Engine,
         *     by calling the endpoint /api/le_enabled. However, this is no
         *     longer available on the new portal, and there is no equivalent
         *     call. So we'll just try to call the LE itself and return false
         *     if we run into any issues. The getClient() will throw an exception
         *     if the key or LE url are invalid.
         */
        try
        {
            getClient().getCourses().get()
            return true
        }
        catch (final Exception $e)
        {
            return false
        }
    }
    
    private LearningEngineClient getClient ()
    {
        final String token = getLearningEngineToken()
        final String url = getLearningEngineURL()
        if (!token)
        {
            throw new IllegalStateException("No valid token provided")
        }
        else if(!url)
        {
            throw new IllegalStateException("No valid URL provided")
        }
        
        return _clients.get("${token} ${url}", {
            return new LearningEngineClient(url, token)
        })
    }
    
    private String getLearningEngineURL ()
    {
        final SettingsProxy settingsProxy = _facade.retrieveProxy(SettingsProxy.NAME)
        final Settings userSettings = settingsProxy.getUserSettings()
        final Settings systemSettings = settingsProxy.getSystemSettings()
                        
        return systemSettings.getString(LESettingNames.LEARNING_ENGINE_URL)
            .or(userSettings.getString(LESettingNames.LEARNING_ENGINE_URL))
            .or(Defaults.LEARNING_ENGINE_URL);
    }
    
    private String getLearningEngineToken ()
    {
        final SettingsProxy settingsProxy = _facade.retrieveProxy(SettingsProxy.NAME)
        final Settings userSettings = settingsProxy.getUserSettings()
        return userSettings.getString(SettingNames.OAUTH_PORTAL_API_KEY).orNull()
    }
    
    private static Node saveLEMetadata (final MediaPackage $mediaPackage, final Map $metadata)
    {
        final Graph graph = $mediaPackage.getGraph().get()        
        Node node = findLearningEngineNode(graph)
        Metadata metadata
        
        if (node)
        {
            metadata = node.config().metadata
        }
        else
        {
            metadata = new Metadata()
            metadata.setID(Identifiers.LEARNING_ENGINE_CONFIG)
            metadata.setName("Learning Engine metadata")
            node = graph.add(PluginNames.METADATA).linkParents(0)
        }

        $metadata.each {key, value ->
            metadata.put(key, value)
        }
        
        node.configure([
            metadata: metadata
        ])
        $mediaPackage.attach(metadata.getID(), metadata)
        graph.save()
        return node
    }
    
    /*
     * Returns the learning engine node in this graph, or null if it does not exist
     */
    private static Node findLearningEngineNode (final Graph $graph)
    {
        return $graph.findNode({
            if (it.config().containsKey("metadata"))
            {
                def metadata = it.config().metadata
                return metadata.getID() == Identifiers.LEARNING_ENGINE_CONFIG
            }
            return false
        } as Predicate).orNull()
    }
}

class LearningEnginePublishEvent extends Event implements MediaPackageEvent
{
    private MediaPackage _mediaPackage;
    
    LearningEnginePublishEvent (MediaPackage $package) 
    {
        super(EventType.PUBLISH_SUCCESS, "${$package.title} was published to the Learning Engine successfully");
        setMediaPackage($package);
    }
    
    public void setMediaPackage(MediaPackage $mediaPackage)
    {
        _mediaPackage = Optional.fromNullable($mediaPackage).or(MediaPackage.Option.INVALID);
        final Map<String,Object> data = new LinkedHashMap<String,Object>(getData());
        data.put("mediaPackageID", _mediaPackage.getID());
        setData(data);
    }
    
    public String getMediaPackageID()
    {
        if (!getData().containsKey("mediaPackageID"))
        {
            return MediaPackage.Option.INVALID.getID();
        }
        
        return getData().get("mediaPackageID").toString();
    }    
}

class LearningEnginePublishFailureEvent extends Event implements MediaPackageEvent
{
    private MediaPackage _mediaPackage;
    private String _error;
    
    LearningEnginePublishFailureEvent (MediaPackage $package, String $error)
    {
        super(EventType.PUBLISH_FAILURE, "${$package.title} was not published to the Learning Engine: " + $error);
        setMediaPackage($package);
        setError($error);
    }
    
    public void setMediaPackage(MediaPackage $mediaPackage)
    {
        _mediaPackage = Optional.fromNullable($mediaPackage).or(MediaPackage.Option.INVALID);
        final Map<String,Object> data = new LinkedHashMap<String,Object>(getData());
        data.put("mediaPackageID", _mediaPackage.getID());
        setData(data);
    }
    
    public void setError (String $error)
    {
        _error = $error;
        final Map<String,Object> data = new LinkedHashMap<String,Object>(getData());
        data.put("error", $error);
        setData(data);
    }
    
    public String getMediaPackageID()
    {
        if (!getData().containsKey("mediaPackageID"))
        {
            return MediaPackage.Option.INVALID.getID();
        }
        
        return getData().get("mediaPackageID").toString();
    }
    
    public String getError ()
    {
        return _error;
    }
}

plugin = new LearningEnginePlugin()